//
// This project using a 2.13" e-Ink Display Module as an example
// For details about this ESP32 module, please visit : https://www.dfrobot.com//product-2101.html
// and https://wiki.dfrobot.com/e-Ink_Display_Module_for_ESP32_SKU_DFR0676
//
// To decode a RF bit stream, we need to know 
// - a start condition
//   A low pulse with duration about 1000us
//  
// - time duration of each bit
//   About 142us, fix duration for each of bit
//  
// - Total number of bits per frame
//   25
//
//  We use a timer with resolution at 100ns for bit duration and start condition measurement. 
//  A RF pin input with edge interrupt generation, on either rising-edge or falling-edge.
//
//

#include <pgmspace.h>
#include "SPI.h"

#include"EPD_Font8x16.h"                        // LCD background logo and fonts data included

#define SPI_MOSI    23                          // SPI Interface for the e-Ink display
#define SPI_MISO    -1
#define SPI_CLK     18

#define ELINK_SS    5                           // Additional hardware connection about the e-Ink display
#define ELINK_BUSY  4
#define ELINK_RESET 16
#define ELINK_DC    17

#define RFIN_PIN    21                          // RF Streams input pin from the RF Module
#define DEBUG_PIN   19                          // a digiral output pin for debug purpose

#define MAX_LINE_BYTES        16                // 16 lines x 8 bits = 128 pixels
#define MAX_COLUMN_BYTES      250               // 250 pixels in column
#define ALLSCREEN_GRAGHBYTES  4000

//
// Defination related to RF bit streams
// All these data were obtained by oscilloscope
// 

#define STREAM_HEAD_MIN     9000                // RF Bit Streams header detection timing : minimum is 9000us
#define STREAM_HEAD_MAX     13000               //   and maximum is 13000us

#define STREAM_BITDURATION  1420                // RF Bit Streams duration, measured as 1360us
#define STREAM_BITDUR_MIN   1200                // Minimum duration for each bit
#define STREAM_BITDUR_MAX   1600                // Maximum duration for each bit
#define STREAM_BITDETECT    600                 // a time delay about 600us after rising-edge detected for a single bit

#define STREAM_BITCOUNT     25                  // Total number of bits per frame
#define STREAM_TIMEOUT      48                  // Number of rising-edge wait when bit duration is not in given specification

//
// Fixed setup data for the e-Ink display
// For details, please refer to https://wiki.dfrobot.com/e-Ink_Display_Module_for_ESP32_SKU_DFR0676
//
const uint8_t LUT_DATA_full[] =
{
  0xA0,  0x90, 0x50, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x50, 0x90, 0xA0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xA0, 0x90, 0x50, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x50, 0x90, 0xA0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

  0x0F, 0x0F, 0x00, 0x00, 0x00,
  0x0F, 0x0F, 0x00, 0x00, 0x03,
  0x0F, 0x0F, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,

  //0x15, 0x41, 0xA8, 0x32, 0x50, 0x2C, 0x0B,
};

//
// Fixed setup data for the e-Ink display
//
const uint8_t LUT_DATA_part[] =
{
  0x40,  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

  0x0A, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00,

  //0x15, 0x41, 0xA8, 0x32, 0x50, 0x2C, 0x0B,
};

//
// Variables related to RF stream decode
//
uint8_t streamState;                            // Decode states, 0: Wait for stream header and 1: decode data bit
uint8_t rfBitCount;                             // Number of data bit read
uint8_t rfDecodeRead;                           // Number of data frame (25-bits) decoded
uint32_t streamCode;                            // a 32-bit dword for saving bit where just read
uint32_t streamDecode[4];                       // an array for saving decoded 25-bit data stream
uint32_t frameTimeout;                          // timeout counter for each frame detection

uint16_t dispTimeOut;                           // timeout counter switch ID display on e-Ink

SPIClass epdSPI(HSPI);
uint8_t SPIsend, SPIreceive;
hw_timer_t *rfTimer = NULL;               

//
// RFIN_PIN edge interrupt serivce routine
// - for start condition detection and data stream timeout 
//
//
void IRAM_ATTR rfDetect_ISR() {
  uint8_t pinState;                             // RF input pin logic level
  uint64_t rfTimerVal;                          // rfDetectTimer counter value, 100ns per count

  timerStop(rfTimer);                           // Stop timer first
  rfTimerVal = timerRead(rfTimer);              // get the counter value
  timerStart(rfTimer);                          // re-start timer
  
  pinState = digitalRead(RFIN_PIN);             // Read the RFIN_PIN logical level
  if (streamState == 0) {                   
    if (pinState == LOW) {                      // if logic level is LOW, that is Falling Edge
      timerWrite(rfTimer, 0L);                  // Reset counter value to ZERO and wait for rising-edge interrupt
    } 
    else {                                      // if logic level is HIGH, that is Rising Edge
                                                // Check the counter value and makesure that it is between 900us and 1300us
                                                //   and total number of frame decoded less than 4
      if ((rfTimerVal >= STREAM_HEAD_MIN) && (rfTimerVal <= STREAM_HEAD_MAX) && (rfDecodeRead < 4)) {
        digitalWrite(DEBUG_PIN, HIGH);          // DEBUG to HIGH, a start condition detected
        timerWrite(rfTimer, 0L);                // Reset counter value to ZERO
        timerAlarmEnable(rfTimer);              // Enable Alarm Interrupt (use timer to decode each bit, delay about 60us after rising-edge)

        streamState ++;                         // streamState = 1 for bit decode
        rfBitCount = 0;                         // Number of bit count to ZERO
        streamCode = 0;                         // Decoded Word to ZERO
      }
      else                                      // If a start condition, the low pulse duration is not match, then ignore it
        streamState = 0;                        // streamState = 0 for start condition detect again
        if (frameTimeout > 0)                   // Check for number of retries.
           frameTimeout --;
        else
          digitalWrite(DEBUG_PIN, HIGH);        // If it is more the 48 times, then DEBUG_PIN change to HIGH for end of frame detection
    }
  }
  else if (streamState == 1) {                  // streamState = 1, for bit decode
    if (pinState == HIGH) {                     // If RFIN_PIN is at logic HIGH, that is rising-edge detected
                                                // If time duration between each rising-edge is between 120us and 160us, then
      if ((rfTimerVal > STREAM_BITDUR_MIN) && (rfTimerVal < STREAM_BITDUR_MAX)) {
        timerWrite(rfTimer, 0L);                // Set timer counter to ZERO
        timerAlarmEnable(rfTimer);              // Enable timer alarm for bit decode
        digitalWrite(DEBUG_PIN, HIGH);          // DEBUG_PIN changed to HIGH state
      }
      else {                                    // otherwise, a wrong bit duration detected
        timerAlarmDisable(rfTimer);             // Disable timer alarm
        timerWrite(rfTimer, 0L);                // Set timer counter to ZERO
        streamState = 0;                        // Restart everything. Now, detects a start condition of a frame
      }
    }
  }
}

//
// This Timer Interrupt Service Routine called by Alarm
// on 60us delay on the rising-edge of RFIN_PIN signal.
//
void IRAM_ATTR rfDetectTimer_ISR() 
{
  digitalWrite(DEBUG_PIN, LOW);			// Put DEBUG_PIN to logic LOW

  streamCode <<= 1;					// Shift left one bit first, if 
  if (digitalRead(RFIN_PIN) == HIGH)            //   the RFIN_PIN is HIGH, then ORed LSB with 1
    streamCode |= 1; 

  rfBitCount ++;						// Add bit received counter by 1, since a bit is received
  if (rfBitCount >= STREAM_BITCOUNT) {		// Stop ALarm service when totally 25 bits received
    timerAlarmDisable(rfTimer);
    timerWrite(rfTimer, 0L);				// streamState go back to ZERO, starts to looking for a start condition
    streamState = 0;

    streamDecode[0] = streamDecode[1];		// Since the Switch ID code repeats four time, we are trying to receive
    streamDecode[1] = streamDecode[2];          // all of them. We need to get at least two of them, to confirm there is 
    streamDecode[2] = streamDecode[3];          // a RF switch is being pressed
    streamDecode[3] = streamCode;
    rfDecodeRead ++;					// No of ID code received add by 1
  }

  frameTimeout = STREAM_TIMEOUT;  			// Reset timeout counter
}

//
// Write Command to EPD
//
void EPD_W21_writeCMD(unsigned char command)
{
  EPD_W21_CheckBUSY();
  epdSPI.beginTransaction(SPISettings(4000000, MSBFIRST, SPI_MODE0));  
  delayMicroseconds(10);
  digitalWrite(ELINK_SS, LOW);
  delayMicroseconds(10);
  digitalWrite(ELINK_DC, LOW);
  SPIreceive = epdSPI.transfer(command); 
  digitalWrite(ELINK_SS, HIGH);
  epdSPI.endTransaction();  
}

//
// Write Data to EPD
//
void EPD_W21_writeDATA(unsigned char data)
{
  EPD_W21_CheckBUSY();
  epdSPI.beginTransaction(SPISettings(4000000, MSBFIRST, SPI_MODE0));  
  delayMicroseconds(10);
  digitalWrite(ELINK_SS, LOW);
  delayMicroseconds(10);
  digitalWrite(ELINK_DC, HIGH);
  delayMicroseconds(10);
  SPIreceive = epdSPI.transfer(data); 
  digitalWrite(ELINK_SS, HIGH);
  epdSPI.endTransaction();  
}

//
// Write a block of Data to EPD
//
void EPD_W21_writeblockDATA(const uint8_t *d, uint16_t size)
{
  uint16_t i;
  EPD_W21_CheckBUSY();
  epdSPI.beginTransaction(SPISettings(4000000, MSBFIRST, SPI_MODE0));  
  delayMicroseconds(10);
  digitalWrite(ELINK_SS, LOW);
  delayMicroseconds(10);
  digitalWrite(ELINK_DC, HIGH);
  delayMicroseconds(10);
  for (i=0; i<size; i++)
    SPIreceive = epdSPI.transfer(*d++); 
  digitalWrite(ELINK_SS, HIGH);
  epdSPI.endTransaction();  
}

//
// Perform RESET for EPD
//
void EPD_W21_Reset(void)
{
  delay(20);
  digitalWrite(ELINK_RESET, LOW);
  delay(20);
  digitalWrite(ELINK_RESET, HIGH);
  delay(200);
}

//
// Check to see if the EPD module is BUSY or NOT
//
void EPD_W21_CheckBUSY(void)
{
  unsigned long start = micros();

  while(1) {
    if (!digitalRead(ELINK_BUSY)) break;
    delay(1);    
    if (micros() - start > 10000000) {
      Serial.println("Busy Timeout!");
      break;
    }
  }
}

//
// Refresh EPD full-screen to update display contents
//
void EPD_Update(void)
{   
  EPD_W21_writeCMD(0x22); 			// Display Update Control 
  EPD_W21_writeDATA(0xF7);   
  EPD_W21_writeCMD(0x20); 			// Activate Display Update Sequence
  EPD_W21_CheckBUSY();

}

//
// Refresh EPD screen partially
//
void EPD_Part_Update(void)
{
  EPD_W21_writeCMD(0x22); 			// Display Update Control 
  EPD_W21_writeDATA(0xFF);   
  EPD_W21_writeCMD(0x20); 			// Activate Display Update Sequence
  EPD_W21_CheckBUSY();    
}

//
// Let EPD display get into sleep mode
//
void EPD_DeepSleep(void)
{  
  EPD_W21_writeCMD(0x10); 			// Enter deep sleep
  EPD_W21_writeDATA(0x01); 
  delay(100);
}


//
// Transfer display data into EPD's display RAM
//
void EPD_SetRAMValue_BaseMap( const unsigned char * datas)
{
  unsigned int i;   
  const unsigned char  *datas_flag;   

  datas_flag=datas;
  EPD_W21_writeCMD(0x24);   			// Write Black and White image to RAM
  for(i=0;i<ALLSCREEN_GRAGHBYTES;i++) {               
    EPD_W21_writeDATA(pgm_read_byte(&datas[i]));
  }

  datas=datas_flag;
  EPD_W21_writeCMD(0x26);   			// Write Black and White image to RAM
  for(i=0;i<ALLSCREEN_GRAGHBYTES;i++) {               
    EPD_W21_writeDATA(pgm_read_byte(&datas[i]));
  }
  EPD_Update();   
}

//
// Fill display RAM with given image
//
void EPD_WhiteScreen_ALL(const unsigned char * data)
{
   unsigned int i;

   EPD_W21_writeCMD(0x24);   			// write RAM for black(0)/white (1)
   for(i=0;i<ALLSCREEN_GRAGHBYTES;i++) {               
     if (data == NULL)
        EPD_W21_writeDATA(0xff);
     else
        EPD_W21_writeDATA(pgm_read_byte(&data[i]));
   }
   EPD_Update();   
}

//
// Partially Fill display RAM with given image
//
void EPD_Dis_Part(unsigned int x_start,unsigned int y_start,const unsigned char * datas,unsigned int PART_COLUMN,unsigned int PART_LINE)
{
  unsigned int i;  
  unsigned int x_end,y_start1,y_start2,y_end1,y_end2;
  x_start=x_start/8;
  x_end=x_start+PART_LINE/8-1; 
  
  y_start1=0;
  y_start2=y_start;
  if(y_start>=256) {
    y_start1=y_start2/256;
    y_start2=y_start2%256;
  }
  y_end1=0;
  y_end2=y_start+PART_COLUMN-1;
  if(y_end2>=256) {
    y_end1=y_end2/256;
    y_end2=y_end2%256;    
  }   
  
  EPD_W21_writeCMD(0x44);       		// set RAM x address start/end, in page 35
  EPD_W21_writeDATA(x_start);    		// RAM x address start at 00h;
  EPD_W21_writeDATA(x_end);    		// RAM x address end at 0fh(15+1)*8->128 
  EPD_W21_writeCMD(0x45);       		// set RAM y address start/end, in page 35
  EPD_W21_writeDATA(y_start2);    		// RAM y address start at 0127h;
  EPD_W21_writeDATA(y_start1);    		// RAM y address start at 0127h;
  EPD_W21_writeDATA(y_end2);    		// RAM y address end at 00h;
  EPD_W21_writeDATA(y_end1);    		// ????=0 


  EPD_W21_writeCMD(0x4E);   			// set RAM x address count to 0;
  EPD_W21_writeDATA(x_start); 
  EPD_W21_writeCMD(0x4F);   			// set RAM y address count to 0X127;    
  EPD_W21_writeDATA(y_start2);
  EPD_W21_writeDATA(y_start1);
  
  
   EPD_W21_writeCMD(0x24);   			// Write Black and White image to RAM
   for(i=0;i<PART_COLUMN*PART_LINE/8;i++) {                         
     EPD_W21_writeDATA(pgm_read_byte(&datas[i]));
   } 
}


//
// Put a string on the EPD module with coordinations
//
void EPD_PrintString(uint8_t sx, uint8_t sy,const char* str, uint8_t isUpdate)
{
  uint8_t slen;
  const char* pt;
  uint16_t sChar;

  pt = str;
  slen = 0;
  while (*pt++)
    slen ++;
  if (slen == 0)
    return;  

  pt = str + slen - 1;
  sx = 249 - sx;            // Maping to LCD's coordinates
  sx = sx - (8 * (slen-1));
  
  while (*pt) {
    sChar = *pt--;
    sChar = sChar - 32;
    sChar = sChar * 16;
    EPD_Dis_Part(sy, sx,&Font8x16[sChar],8,16);
    sx = sx + 8;
  }

  if (isUpdate)
    EPD_Part_Update();
}


//
// Print a 32-bit HEX value on EPD, with given coordinations
//
void EPD_PrintHex32bit(uint8_t hx, uint8_t hy, uint32_t hex, uint8_t isUpdate)
{
  uint16_t hChar, i;

  hx = 249 - hx;            // Maping to LCD's coordinates
  if (hx > 56)
    hx = hx - 56;

  for (i=0; i<8; i++) {
    hChar = (hex & 0x0000000F);
    if (hChar <= 9) {
      hChar = hChar + '0';
    }
    else {
      hChar = hChar - 10;
      hChar = hChar + '0';
      hChar = hChar + 17;
    }

    hChar = hChar - 32;
    hChar = hChar * 16;
    EPD_Dis_Part(hy, hx,&Font8x16[hChar],8,16);
    hx = hx + 8;
    hex >>= 4;
  }

  if (isUpdate)
    EPD_Part_Update();
}


//
// Standard Initial Code for EPD Module
//
void EPD_init(void)
{
  EPD_W21_Reset(); 			// Electronic paper IC reset

  EPD_W21_CheckBUSY();   
  EPD_W21_writeCMD(0x12);  		// SWRESET
  EPD_W21_CheckBUSY();   
    
  EPD_W21_writeCMD(0x01); 		// Driver output control      
  EPD_W21_writeDATA(0xF9);
  EPD_W21_writeDATA(0x00);
  EPD_W21_writeDATA(0x00);

  EPD_W21_writeCMD(0x11); 		// Data entry mode       
  EPD_W21_writeDATA(0x01);

  EPD_W21_writeCMD(0x44); 		// Set Ram-X address start/end position   
  EPD_W21_writeDATA(0x00);
  EPD_W21_writeDATA(0x0F);    	// 0x0F-->(15+1)*8=128

  EPD_W21_writeCMD(0x45); 		// Set Ram-Y address start/end position          
  EPD_W21_writeDATA(0xF9);   		// 0xF9-->(249+1)=250
  EPD_W21_writeDATA(0x00);
  EPD_W21_writeDATA(0x00);
  EPD_W21_writeDATA(0x00); 

  EPD_W21_writeCMD(0x3C); 		// BorderWavefrom
  EPD_W21_writeDATA(0x01);  
      
  EPD_W21_writeCMD(0x18); 		// Reading temperature sensor
  EPD_W21_writeDATA(0x80);  

  EPD_W21_writeCMD(0x4E);   		// Set RAM x address count to 0;
  EPD_W21_writeDATA(0x00);
  EPD_W21_writeCMD(0x4F);   		// Set RAM y address count to 0X199;    
  EPD_W21_writeDATA(0xF9);
  EPD_W21_writeDATA(0x00);
  EPD_W21_CheckBUSY();  
}


//
// Arduino Setup()
//
void setup() {
  // put your setup code here, to run once:
  streamState = 0;						// Set streamState to ZERO, to looking for a start condition
  frameTimeout = STREAM_TIMEOUT;          		// Reset timeout Counter
  dispTimeOut = 0;                        		// Reset display timeout counter

  Serial.begin(115200);						// Enable Serial communication for debug purpose
  Serial.println();
  Serial.println("setup");

  pinMode(ELINK_BUSY, INPUT_PULLUP);			// Config I/O pins for EPD Module
  digitalWrite(ELINK_SS, HIGH);
  digitalWrite(ELINK_RESET, HIGH);
  digitalWrite(ELINK_DC, HIGH);
  pinMode(ELINK_SS, OUTPUT);
  pinMode(ELINK_DC, OUTPUT);
  pinMode(ELINK_RESET, OUTPUT);

  pinMode(DEBUG_PIN, OUTPUT);					// DEBUG_PIN as output 
  digitalWrite(DEBUG_PIN, LOW);				//   with Logic LOW state

  EPD_W21_Reset(); 						// Electronic paper IC reset
  epdSPI.begin(SPI_CLK, SPI_MISO, SPI_MOSI, ELINK_SS);

  EPD_init();  							// Electronic paper initialization
  EPD_SetRAMValue_BaseMap(gImage_basemap);  		// Show background image
  delay(100);
  EPD_PrintString(0, 72, "RF SWITCH ID:", false);
  EPD_PrintHex32bit(112, 72, 0, true);

  Serial.println("Initial Done.");

  // Enable Timer0 with prescaler 8, count up
  // So the resolution of counter0 is 100ns
  //
  rfTimer = timerBegin(0, 8, true);
  timerAttachInterrupt(rfTimer, &rfDetectTimer_ISR, true);
  timerAlarmWrite(rfTimer, STREAM_BITDETECT, false);
  timerAlarmDisable(rfTimer);  

  pinMode(RFIN_PIN, INPUT_PULLUP);				// RFIN_PIN configed as input pin
  attachInterrupt(RFIN_PIN, rfDetect_ISR, CHANGE);	//   and enable edge detection interrupt
}

//
// Arduino main loop()
//
void loop() {
  // put your main code here, to run repeatedly:
  uint16_t i, count;
  uint32_t code; 
  
  if ((rfDecodeRead >= 1) && (frameTimeout == 0)) {
    rfDecodeRead = 0;
    frameTimeout = STREAM_TIMEOUT;

    count = 0;
    for (i=0; i<4; i++) {
      Serial.print(streamDecode[i], HEX);
      Serial.print(' ');

      if (streamDecode[i] != 0)
        code = streamDecode[i];

      if (code == streamDecode[i])
        count ++;  
      
      streamDecode[i] = 0;
    }
    Serial.println();

    if (count > 1) { 
      EPD_PrintString(0, 72, "RF SWITCH ID:", false);
      EPD_PrintHex32bit(112, 72, code, true);
      dispTimeOut = 100;
    }
  }

  if (dispTimeOut > 0) {
    dispTimeOut = dispTimeOut - 1;
    if (dispTimeOut == 0) {
      EPD_PrintString(0, 72, "RF SWITCH ID:", false);
      EPD_PrintHex32bit(112, 72, 0, true);
    }
  }

  delay(10);

}
